# %% imports
from autogaita.resources.utils import bin_num_to_percentages, write_issues_to_textfile
from autogaita.group.group_utils import save_figures
import os
import string
import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler
from sklearn.decomposition import PCA
import openpyxl
import seaborn as sns
import matplotlib.pyplot as plt
from matplotlib.animation import FuncAnimation, FFMpegWriter
import pdb

# %% constants
from autogaita.resources.constants import (
    ID_COL,
    SC_PERCENTAGE_COL,
)
from autogaita.group.group_constants import (
    GROUP_COL,
    PCA_BARPLOT_BARCOLOR,
    PCA_BARPLOT_LINECOLOR,
    PCA_CUSTOM_SCATTER_OUTER_SEPARATOR,
    PCA_CUSTOM_SCATTER_INNER_SEPARATOR,
)

# %% ...........................  workflow step #3 - PCA  ..............................


def PCA_main(avg_dfs, folderinfo, cfg, plot_panel_instance):
    """PCA on features of interest"""

    # unpack
    PCA_custom_scatter_PCs = cfg["PCA_custom_scatter_PCs"]
    PCA_bins = cfg["PCA_bins"]

    # convert PCA_bins string to list of ints
    if PCA_bins:
        cfg = convert_PCA_bins_to_list(folderinfo, cfg)
        info_string = "Desired PCA bins applied to cycle-bins resulted in:\n|"
        for bin in cfg["PCA_bins"]:
            info_string += str(bin) + "|"
        print(info_string)
    # create the input dataframe
    PCA_df, features = create_PCA_df(avg_dfs, folderinfo, cfg)
    # run the PCA
    PCA_df, PCA_info = run_PCA(PCA_df, features, cfg)
    # save PCA info to xlsx files
    PCA_info_to_xlsx(PCA_df, PCA_info, folderinfo, cfg)
    if PCA_info["number_of_PCs"] < 2:
        PCA_error_message_1 = (
            "\n***********\n! ERROR !\n***********\n"
            + "PC scatters were not generated because there are less than 2 PCs!"
            + "\nThis hints at a low value of PCA_n_components using the "
            + "var-explained approach!"
        )
        print(PCA_error_message_1)
        write_issues_to_textfile(PCA_error_message_1, folderinfo)
        return
    # plot barplot of cumulative explained variance
    plot_PCA_barplot(PCA_info, folderinfo, cfg, plot_panel_instance)
    # plot the standard scatterplot
    plot_PCA_scatterplots(
        "Standard", PCA_df, PCA_info, folderinfo, cfg, plot_panel_instance
    )
    # plot custom scatterplots if user wanted
    # => loop over outer separator of PC-duos/trios
    if PCA_custom_scatter_PCs:
        for i in range(
            len(PCA_custom_scatter_PCs.split(PCA_CUSTOM_SCATTER_OUTER_SEPARATOR))
        ):
            plot_PCA_scatterplots(
                "Custom #" + str((i + 1)),
                PCA_df,
                PCA_info,
                folderinfo,
                cfg,
                plot_panel_instance,
            )


def convert_PCA_bins_to_list(folderinfo, cfg):
    """Create PCA bin variable (list of to-be-included bin_num ints) from user input
    (string).
    Approach
    --------
    1. (Note we already removed spaces in group_1_prep). Also initialise a string list that has all bins (, separated) and an empty list
    2. Iterate over string list and extract start and end indices if it's a range (idenfying range by "-")
    3. Concatenate the range to the int list, make sure that you add a 1 to end idx
    to have user input be inclusive
    4. If it's not a range, just add the int to the int list
    5. Now remove all values that are not part of our SC Percentages (use bin_num)
    for this
    """
    # unpack
    PCA_bins = cfg["PCA_bins"]
    bin_num = cfg["bin_num"]

    PCA_bin_str_list = PCA_bins.split(",")
    PCA_bin_int_list = []
    for bin in PCA_bin_str_list:
        if "-" in bin:
            bin_range = bin.split("-")
            # +1 in indexing of bin_range [1] ensures including user input fully!
            try:
                PCA_bin_int_list = np.concatenate(
                    [
                        PCA_bin_int_list,
                        np.arange(int(bin_range[0]), int(bin_range[1]) + 1),
                    ]
                )
            except ValueError:
                PCA_bins_error_message = (
                    "\n*********\n! ERROR !\n*********\n"
                    + "\nPCA bin range input invalid."
                    + "\nSeems like you used a hyphen without numbers around it."
                    + "\nPlease check your input and try again!"
                )
                print(PCA_bins_error_message)
                write_issues_to_textfile(PCA_bins_error_message, folderinfo)
                raise ValueError(PCA_bins_error_message)

        else:
            PCA_bin_int_list = np.concatenate([PCA_bin_int_list, [int(bin)]])
        PCA_bin_int_list = PCA_bin_int_list.astype(int)  # for intersect below
    # Note that we can be sure that this is the same list that was used at the first
    # level since it's the same function used there
    percentages_list = bin_num_to_percentages(bin_num)
    # Now using intersect 1d that returns the sorted unique values that are in both
    # arrays
    cfg["PCA_bins"] = np.intersect1d(PCA_bin_int_list, percentages_list)
    # Just in case... remove everything over 100 (our error above ensures nothing
    # smaller than 0 already)
    cfg["PCA_bins"] = cfg["PCA_bins"][cfg["PCA_bins"] <= 100]
    return cfg


def create_PCA_df(avg_dfs, folderinfo, cfg):
    """Create a ID x ID_COL + features dataframe to be used by PCA"""

    # unpack
    group_names = folderinfo["group_names"]
    PCA_vars = cfg["PCA_variables"]
    bin_num = cfg["bin_num"]
    PCA_bins = cfg["PCA_bins"]

    PCA_df = pd.DataFrame(data=None)
    # create a list of features for series & dfs (features are vars @ each SC % bin)
    features = []
    for var in PCA_vars:
        if len(PCA_bins) > 0:
            for bin in PCA_bins:
                features.append(var + " " + str(bin))
        else:
            for b in range(bin_num):
                bin_in_percent = int(((1 + b) / bin_num) * 100)
                features.append(var + " " + str(bin_in_percent))

    # for each mouse, create a series to concat to PCA_df
    for g, group_name in enumerate(group_names):
        for ID in pd.unique(avg_dfs[g][ID_COL]):
            this_list = [group_name, ID]
            if len(PCA_bins) > 0:  # empty strings are len = 0
                # first line gives True and False but we need indexes so it's
                # compatible with my (previous) approach of using iloc and lists
                row_idxs = (avg_dfs[g][ID_COL] == ID) & (
                    avg_dfs[g][SC_PERCENTAGE_COL].isin(PCA_bins)
                )
                row_idxs = np.where(row_idxs == True)[0]
            else:
                row_idxs = np.where(avg_dfs[g][ID_COL] == ID)[0]
            for var in PCA_vars:
                joint_col_idx = avg_dfs[g].columns.get_loc(var)
                # this data is of current ID or IDxSC% combo and current feature
                this_data = list(avg_dfs[g].iloc[row_idxs, joint_col_idx].values)
                this_list.extend(this_data)  # so we have ID_COL as 1st value
            this_series = pd.Series(this_list)
            if PCA_df.empty:
                PCA_df = pd.DataFrame(this_series).transpose()
            else:
                # transpose series, transform to df and concat to row-axis
                PCA_df = pd.concat([PCA_df, this_series.to_frame().T], axis=0)
    # this check is me being paranoid but whatever let's just leave it in
    if len(PCA_df.columns) != len(features) + 2:
        critical_PCA_df_error_message = (
            "\n*********\n! ERROR !\n*********\n"
            + "PCA_df creation failed since columns did not match features!\n"
            + "This should never happen, please contact me!"
        )
        print(critical_PCA_df_error_message)
        write_issues_to_textfile(critical_PCA_df_error_message, folderinfo)
        raise ValueError(critical_PCA_df_error_message)
    # add colnames after the last mouse (makes concat'ing series 2 df easier)
    PCA_df.columns = [GROUP_COL] + [ID_COL] + features
    return PCA_df, features


def run_PCA(PCA_df, features, cfg):
    """Runs the PCA on a limb's feature (e.g. y or z coordinates)"""

    # unpack
    PCA_n_components = cfg["PCA_n_components"]

    # run
    if 0 < PCA_n_components < 1:  # n_components until % explained var.
        PCA_model = PCA(n_components=PCA_n_components, svd_solver="full")
    else:
        PCA_model = PCA(n_components=int(PCA_n_components))
    x = PCA_df.loc[:, features]
    # standardise here so that EACH FEATURE has mean=0 & std=1
    # => you can check this with np.mean/std
    # => you can also check that this operates on columns (features) and not rows (IDs)
    x = StandardScaler().fit_transform(x)
    # now create a df with the standardised features to add features to it (this gives
    # us the feature_names_in attribute of PCA_model which we will use for PCA info
    # excel files)
    x = pd.DataFrame(x, columns=features)
    # NOTE that this next line changes the PCA_model variable!
    PCs = PCA_model.fit_transform(x)
    # NOTE! we use number_of_PCs from here onwards (in info/plots/etc) to differentiate
    #       between n_components which can be smaller than 1!
    number_of_PCs = np.shape(PCs)[1]
    for i in range(number_of_PCs):
        PCA_df["PC " + str(i + 1)] = PCs[:, i]
    PCA_info = {
        "features": PCA_model.feature_names_in_,
        "number_of_PCs": number_of_PCs,
        "explained_vars": PCA_model.explained_variance_ratio_,
        "eigenvectors": PCA_model.components_,
    }
    return PCA_df, PCA_info


def PCA_info_to_xlsx(PCA_df, PCA_info, folderinfo, cfg):
    """Save three PCA info excel files:
    1) explained_var & eigenvectors of PCs
    2) ID x PC values
    3) top 20 features for each PC using eigenvectors
    """

    # unpack
    results_dir = folderinfo["results_dir"]
    features = PCA_info["features"]  # PCA info!
    number_of_PCs = PCA_info["number_of_PCs"]
    explained_vars = PCA_info["explained_vars"]
    eigenvectors = PCA_info["eigenvectors"]

    # 1) PCs' explained variances & eigenvectors

    #  initialise
    workbook = openpyxl.Workbook()
    sheet = workbook.active
    sheet.title = "PCA Info"

    # add column headers
    for pc in range(number_of_PCs):
        # pdb.set_trace()
        sheet.cell(row=1, column=pc + 2, value="PC " + str(pc + 1))
    # add cell values: explained variance
    sheet.cell(row=2, column=1, value="Explained Var. (%)")
    for pc in range(number_of_PCs):
        sheet.cell(row=2, column=pc + 2, value=round(explained_vars[pc] * 100, 2))
    # add cell values: eigenvectors
    sheet.cell(row=4, column=1, value="Features")
    for i in range(len(features)):
        # row i + 5 because excel starts counting at 1 and: row2=var_exp/row3=empty/
        # row5=feature heading
        sheet.cell(row=i + 5, column=1, value=features[i])
        for pc in range(number_of_PCs):
            # column is pc+2 because we want pc=0 to be in xlsx column 2 etc.
            sheet.cell(row=i + 5, column=pc + 2, value=eigenvectors[pc, i])
    # save
    workbook.save(os.path.join(results_dir, "PCA Info.xlsx"))

    # 2) ID x PC value information
    PCA_ID_info_df = PCA_df[[GROUP_COL, ID_COL] + list(PCA_df.columns[-number_of_PCs:])]
    PCA_ID_info_df.to_excel(os.path.join(results_dir, "PCA ID Info.xlsx"), index=False)

    # 3) Find the biggest 20 values of eigenvectors for each PC
    if len(features) < 20:
        top_number = len(features)
    else:
        top_number = 20
    top_features_df = pd.DataFrame()
    for pc in range(number_of_PCs):
        # argsort returns idxs that would sort in ascending order, so we reverse it
        # (::-1)
        top_features = np.argsort(np.abs(eigenvectors[pc]))[::-1][:top_number]
        top_values = eigenvectors[pc, top_features]
        top_features_df["PC " + str(pc + 1) + " Features"] = features[top_features]
        top_features_df["PC " + str(pc + 1) + " Value"] = top_values
    filename = "PCA Feature Summary.xlsx"
    top_features_df.to_excel(os.path.join(results_dir, filename), index=False)


def plot_PCA_barplot(PCA_info, folderinfo, cfg, plot_panel_instance):
    """Plot a barplot illustrating PCs explained variance"""

    # unpack
    number_of_PCs = PCA_info["number_of_PCs"]  # PCA info!
    explained_vars = PCA_info["explained_vars"]
    results_dir = folderinfo["results_dir"]
    dont_show_plots = cfg["dont_show_plots"]

    # plot
    df = pd.DataFrame(
        data=explained_vars * 100,
        columns=["Explained Variance"],
        index=np.arange(1, PCA_info["number_of_PCs"] + 1),
    )
    cumulative_explained_var = []
    for i in range(number_of_PCs):
        cumulative_explained_var.append(np.sum(explained_vars[: i + 1]) * 100)
    f, ax = plt.subplots(1, 1)
    sns.barplot(df["Explained Variance"], color=PCA_BARPLOT_BARCOLOR, ax=ax)
    ax.set_xlabel("Principal Components")
    ax.set_ylabel("Explained Variance (%)")
    ax.plot(
        cumulative_explained_var,
        color=PCA_BARPLOT_LINECOLOR,
        linestyle=":",
        marker="o",
        label="Cumulative Explained Variance",
    )
    info_string = "PCA Explained Variance"
    save_figures(f, results_dir, info_string)
    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)


def plot_PCA_scatterplots(
    which_plot, PCA_df, PCA_info, folderinfo, cfg, plot_panel_instance
):
    """Plot a scatterplot and colour based on group name"""

    # unpack
    group_names = folderinfo["group_names"]
    results_dir = folderinfo["results_dir"]
    group_color_dict = cfg["group_color_dict"]
    PCA_custom_scatter_PCs = cfg["PCA_custom_scatter_PCs"]
    PCA_save_3D_video = cfg["PCA_save_3D_video"]
    dont_show_plots = cfg["dont_show_plots"]
    legend_outside = cfg["legend_outside"]

    # number of PCs & custom flag to control standard/custom plots
    number_of_PCs = PCA_info["number_of_PCs"]  # PCA info!
    custom_flag = False
    if "Custom" in which_plot:
        custom_flag = True
        # first, extract PC numbers from the cfg string and test input validity
        # => important: we need to subtract 1 after extracting from the string because
        #               PC numbers are 1-indexed
        PC_idx = int(which_plot[-1]) - 1  # here!
        all_scatter_PC_sets = PCA_custom_scatter_PCs.split(
            PCA_CUSTOM_SCATTER_OUTER_SEPARATOR
        )
        this_scatter_PC_set = all_scatter_PC_sets[PC_idx]
        scatter_PC_nums = this_scatter_PC_set.split(PCA_CUSTOM_SCATTER_INNER_SEPARATOR)
        scatter_PC_nums = [int(this_PC) for this_PC in scatter_PC_nums]
        # Important Note!
        # => below test must be done on "real" number of PCs of full PCA from PCA info
        if np.max(scatter_PC_nums) > number_of_PCs:
            invalid_custom_PC_scatter_message = (
                "\n*********\n! ERROR !\n*********\n"
                + "\nCustom PCA scatterplot configuration invalid for:\n"
                + PCA_custom_scatter_PCs[PC_idx]
                + "!\nThe PC you wanted was not computed by PCA!"
            )
            print(invalid_custom_PC_scatter_message)
            write_issues_to_textfile(invalid_custom_PC_scatter_message, folderinfo)
            return
        # after the test we can overwrite "real" number of PCs of full PCA here since
        # we only use this value to know if we also do a 3D plot or not from here on
        number_of_PCs = len(scatter_PC_nums)

    # prepare PC strings based on which plot we are making and extract col_idxs from
    # PCA_df for scatter
    if custom_flag:
        PC1_str = "PC " + str(scatter_PC_nums[0])
        PC2_str = "PC " + str(scatter_PC_nums[1])
        if number_of_PCs > 2:
            PC3_str = "PC " + str(scatter_PC_nums[2])
    else:
        PC1_str = "PC 1"
        PC2_str = "PC 2"
        if number_of_PCs > 2:
            PC3_str = "PC 3"
    PC1_col_idx = PCA_df.columns.get_loc(PC1_str)
    PC2_col_idx = PCA_df.columns.get_loc(PC2_str)
    if number_of_PCs > 2:
        PC3_col_idx = PCA_df.columns.get_loc(PC3_str)

    # loop over groups, and call scatter 3x so we can have a correct legend
    f, ax = plt.subplots(1, 1)
    if number_of_PCs > 2:
        f_3d = plt.figure()
        ax_3d = f_3d.add_subplot(111, projection="3d")
    for g, group_name in enumerate(group_names):
        row_idxs = np.where(PCA_df[GROUP_COL] == group_name)[0]
        x = PCA_df.iloc[row_idxs, PC1_col_idx].values
        y = PCA_df.iloc[row_idxs, PC2_col_idx].values
        ax.scatter(x, y, color=group_color_dict[group_name], label=group_name)
        if number_of_PCs > 2:
            z = PCA_df.iloc[row_idxs, PC3_col_idx].values
            ax_3d.scatter(
                x, y, z, color=group_color_dict[group_name], s=60, label=group_name
            )
    # legend, labels & info-string for title/save
    if legend_outside is True:
        ax.legend(group_names, loc="center left", bbox_to_anchor=(1, 0.5))
    elif legend_outside is False:
        ax.legend(group_names)
    ax.set_xlabel(PC1_str)
    ax.set_ylabel(PC2_str)
    info_string = which_plot + " 2D PCA Scatterplot"
    ax.set_title(info_string)
    if number_of_PCs > 2:  # 3d scatterplot
        ax_3d.view_init(30, 125)
        # legend adjustments
        if legend_outside is True:
            ax_3d.legend(group_names, loc="center right", bbox_to_anchor=(0, 0.5))
        elif legend_outside is False:
            ax_3d.legend(group_names)
        ax_3d.set_xlabel(PC1_str)
        ax_3d.set_ylabel(PC2_str)
        ax_3d.set_zlabel(PC3_str)
        info_string_3d = which_plot + " 3D PCA Scatterplot"
        ax_3d.set_title(info_string_3d)
    save_figures(f, results_dir, info_string)
    # add figure to plot panel figures list
    if dont_show_plots is False:  # -> show plot panel
        plot_panel_instance.figures.append(f)

    # 3d scatterplot image file
    if number_of_PCs > 2:
        save_figures(f_3d, results_dir, info_string_3d)
        # add figure to plot panel figures list
        if dont_show_plots is False:  # -> show plot panel
            plot_panel_instance.figures.append(f_3d)

        # 3d scatterplot rotating video file
        if PCA_save_3D_video:

            def init():  # create the animation
                return (f_3d,)

            def animate(frame):
                ax_3d.view_init(elev=10, azim=frame)
                return (f_3d,)

            anim = FuncAnimation(
                f_3d, animate, frames=np.arange(0, 360, 1), interval=20, blit=True
            )
            writervideo = FFMpegWriter(fps=30)  # save to m4 using ffmpeg writer
            anim.save(
                os.path.join(results_dir, info_string_3d + ".mp4"), writer=writervideo
            )



from autogaita.dlc.dlc_1_preparation import some_prep
from autogaita.dlc.dlc_2_sc_extraction import extract_stepcycles
from autogaita.common2D.common2D_3_analysis import (
    analyse_and_export_stepcycles,
    add_step_separators,
    add_angles,
    add_x_velocities,
    add_angular_velocities,
    standardise_x_y_and_add_features_to_one_step,
)
from hypothesis import given
import hypothesis.strategies as st
from hypothesis.extra.numpy import arrays
import os
import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest
import pdb


# %%................................  fixtures  ........................................
@pytest.fixture
def extract_data_using_some_prep(extract_info, extract_folderinfo, extract_cfg):
    data = some_prep(extract_info, extract_folderinfo, extract_cfg)
    return data


@pytest.fixture
def extract_info(tmp_path):
    info = {}
    info["mouse_num"] = 15
    info["run_num"] = 3
    info["name"] = "ID " + str(info["mouse_num"]) + " - Run " + str(info["run_num"])
    info["results_dir"] = os.path.join(tmp_path, info["name"])
    return info


@pytest.fixture
def extract_folderinfo():
    folderinfo = {}
    folderinfo["root_dir"] = "tests/test_data/dlc_data"
    folderinfo["sctable_filename"] = (
        "correct_annotation_table.xlsx"  # has to be an excel file
    )
    folderinfo["data_string"] = "SIMINewOct"
    folderinfo["beam_string"] = "BeamTraining"
    folderinfo["premouse_string"] = "Mouse"
    folderinfo["postmouse_string"] = "25mm"
    folderinfo["prerun_string"] = "run"
    folderinfo["postrun_string"] = "6DLC"
    return folderinfo


@pytest.fixture
def extract_cfg():
    cfg = {}
    cfg["sampling_rate"] = 100
    cfg["subtract_beam"] = True
    cfg["dont_show_plots"] = True
    cfg["convert_to_mm"] = True
    cfg["pixel_to_mm_ratio"] = 3.76
    cfg["x_sc_broken_threshold"] = 200  # optional cfg
    cfg["y_sc_broken_threshold"] = 50
    cfg["x_acceleration"] = True
    cfg["angular_acceleration"] = True
    cfg["save_to_xls"] = True
    cfg["bin_num"] = 25
    cfg["plot_SE"] = True
    cfg["standardise_y_at_SC_level"] = False
    cfg["standardise_y_to_a_joint"] = True
    cfg["y_standardisation_joint"] = ["Knee"]  # "Hind paw tao"]
    cfg["plot_joint_number"] = 3
    cfg["color_palette"] = "viridis"
    cfg["legend_outside"] = True
    cfg["invert_y_axis"] = True
    cfg["flip_gait_direction"] = True
    cfg["analyse_average_x"] = True
    cfg["standardise_x_coordinates"] = True
    cfg["x_standardisation_joint"] = ["Hind paw tao"]
    cfg["hind_joints"] = ["Hind paw tao", "Ankle", "Knee", "Hip", "Iliac Crest"]
    cfg["fore_joints"] = [
        "Front paw tao ",
        "Wrist ",
        "Elbow ",
        "Lower Shoulder ",
        "Upper Shoulder ",
    ]
    cfg["beam_col_left"] = ["BeamLeft"]  # BEAM_COL_LEFT & _RIGHT must be lists of len=1
    cfg["beam_col_right"] = ["BeamRight"]
    cfg["beam_hind_jointadd"] = ["Tail base ", "Tail center ", "Tail tip "]
    cfg["beam_fore_jointadd"] = ["Nose ", "Ear base "]
    cfg["angles"] = {
        "name": ["Ankle ", "Knee ", "Hip "],
        "lower_joint": ["Hind paw tao ", "Ankle ", "Knee "],
        "upper_joint": ["Knee ", "Hip ", "Iliac Crest "],
    }
    return cfg


# %%.......  main analysis: sc-lvl y-norm, features, df-creation & export ..........


def test_height_standardisation_no_beam():
    """Unit test normalising heights if there is no beam"""
    # Create a sample DataFrame
    data = {
        "first_y": [-10, -5, 0, 10],
        "second_y": [10, 20, 30, 40],
        "third_y": [5, 15, 25, 35],
    }
    step = pd.DataFrame(data)
    y_cols = [col for col in step.columns if col.endswith("y")]
    this_y_min = step[y_cols].min().min()
    step[y_cols] -= this_y_min
    # Expected result
    expected_result = {
        "first_y": [0, 5, 10, 20],
        "second_y": [20, 30, 40, 50],
        "third_y": [15, 25, 35, 45],
    }
    expected_step = pd.DataFrame(expected_result)
    # Compare the result with the expected result
    pdt.assert_frame_equal(step, expected_step)


cases = [
    ((0, 0), (1, 0), (0, 1), 90), 
    ((0, 0), (1, 0), (0, 0.5), 90),
    ((0, 0), (1, 0), (2, 0), 0)
]  # fmt: skip
@pytest.mark.parametrize("angle_x_y, lower_x_y, upper_x_y, expected_angle", cases)
def test_angles(angle_x_y, lower_x_y, upper_x_y, expected_angle):
    step = (
        pd.Series(
            {
                "angle x": angle_x_y[0],
                "angle y": angle_x_y[1],
                "lower x": lower_x_y[0],
                "lower y": lower_x_y[1],
                "upper x": upper_x_y[0],
                "upper y": upper_x_y[1],
            }
        )
        .to_frame()
        .T
    )

    cfg = {}
    cfg["angles"] = {
        "name": ["angle "],
        "lower_joint": ["lower "],
        "upper_joint": ["upper "],
    }
    step = add_angles(step, cfg)
    assert step["angle Angle"].values == expected_angle


def test_angles_not_depending_on_x_coordinate_standardisation(
    extract_info, extract_folderinfo, extract_cfg
):
    data = some_prep(extract_info, extract_folderinfo, extract_cfg)
    all_cycles = extract_stepcycles(data, extract_info, extract_folderinfo, extract_cfg)
    results = analyse_and_export_stepcycles(data, all_cycles, extract_info, extract_cfg)
    all_steps_data = results["all_steps_data"]
    x_standardised_steps_data = results["x_standardised_steps_data"]
    angle_cols = [col for col in all_steps_data.columns if col.endswith("Angle")]
    for angle_col in angle_cols:
        pdt.assert_series_equal(
            all_steps_data[angle_col], x_standardised_steps_data[angle_col]
        )


@given(x_y_coordinates = st.tuples(st.one_of(st.integers(), st.floats()), st.one_of(st.integers(), st.floats())))  # fmt: skip
@pytest.mark.filterwarnings("ignore:invalid value encountered")
def test_angles_is_nan_when_all_points_are_at_same_location(x_y_coordinates):
    angle_x_y = lower_x_y = upper_x_y = x_y_coordinates
    step = (
        pd.Series(
            {
                "angle x": angle_x_y[0],
                "angle y": angle_x_y[1],
                "lower x": lower_x_y[0],
                "lower y": lower_x_y[1],
                "upper x": upper_x_y[0],
                "upper y": upper_x_y[1],
            }
        )
        .to_frame()
        .T
    )

    cfg = {}
    cfg["angles"] = {
        "name": ["angle "],
        "lower_joint": ["lower "],
        "upper_joint": ["upper "],
    }
    step = add_angles(step, cfg)
    assert step["angle Angle"].isna().all()


def test_velocities():
    """Unit test of how velocities are added
    A Note
    ------
    If you ever consider testing if velocities are expected after flipping x-values using flip_mouse_body - DON'T!
    => We only flip if gait direction is right=>left in the first place and thus: velocities are CORRECT and comparable to non-flipped left=>right videos.
    => NU - write a test that confirms this formally
    """
    data = {
        "Sample x": [0.0, 2.0, 4.0, 8.0, 8.0, 4.0, 2.0, 0.0, -2.0, -10.0],
        "Sample Angle": [-10.0, -8.0, -7.0, -4.0, 0.0, 4.0, 10.0, 20.0, 10.0, 0.0],
    }
    step = pd.DataFrame(data)
    cfg = {
        "hind_joints": ["Sample "],
        "x_acceleration": True,
        "angular_acceleration": True,
    }
    step = add_x_velocities(step, cfg)
    step = add_angular_velocities(step, cfg)
    # expected values were obtained by calling np.gradient on arrays above
    expected_values = {
        "Sample Velocity": [2.0, 2.0, 3.0, 2.0, -2.0, -3.0, -2.0, -2.0, -5.0, -8.0],
        "Sample Acceleration": [0.0, 0.5, 0.0, -2.5, -2.5, 0.0, 0.5, -1.5, -3.0, -3.0],
        "Sample Angle Velocity": [2.0, 1.5, 2.0, 3.5, 4.0, 5.0, 8.0, 0.0, -10.0, -10.0],
        "Sample Angle Acceleration": [
            -0.5,
            0.0,
            1.0,
            1.0,
            0.75,
            2.0,
            -2.5,
            -9.0,
            -5.0,
            0.0,
        ],
    }
    for key in expected_values.keys():
        assert all(expected_values[key] == step[key])


def test_step_separators():
    """Unit test of how step separators are added to df"""
    df_no_separators = pd.DataFrame(
        {"A": [1, 2, 3, 4, 5], "B": ["yes", "no", "yes", "no", "yes"]}
    )
    nanvector = df_no_separators.loc[[1]]
    nanvector[:] = np.nan
    numvector = df_no_separators.loc[[1]]
    numvector[:] = 1
    df_manually_added_separators = pd.DataFrame(
        {
            "A": [1, 2, 3, 4, 5, np.nan, 1, np.nan],
            "B": ["yes", "no", "yes", "no", "yes", np.nan, 1, np.nan],
        }
    )
    df_manually_added_separators.index = [0, 1, 2, 3, 4, 1, 1, 1]
    df_func_added_separators = add_step_separators(
        df_no_separators, nanvector, numvector
    )
    pd.testing.assert_frame_equal(
        df_func_added_separators, df_manually_added_separators
    )



from autogaita import group
from autogaita.group.group_constants import STATS_TXT_FILENAME
import pandas as pd
import pandas.testing as pdt
import filecmp
import os
import shutil
import pdb
import pytest

# ...........................  GROUP APPROVAL TESTS STRUCTURE  .........................
# 1. Run autogaita.group for example group (3 beams) data (with the cfg used there)
#    store results in a temporary path using tmp_path input.
# 2. Load the "Grand Average Stepcycles".xlsx & "Grand Standard Devs. Stepcycle.xlsx"
#    files from the repo (TRUE PATH) - test for equivalence with TEST PATH
# 3. Then test if PCA XLS file is equal
# 4. Finally test if Stats.txt files are equal

# A Note
# ------
# We don't test the cluster extent test in an automated way.
# We have results of simulations that ran for over a week in our preprint.
# If the cluster extent test should change for some reason - re-run those simulations.


# ...............................  PREPARE - THREE FIXTURES   ..........................


@pytest.fixture
def extract_true_dir():
    return "example data/group/"


@pytest.fixture
def extract_folderinfo(tmp_path):
    folderinfo = {}
    folderinfo["group_names"] = ["5 mm", "12 mm", "25 mm"]
    folderinfo["group_dirs"] = [
        "example data/5mm/Results/",
        "example data/12mm/Results/",
        "example data/25mm/Results/",
    ]
    folderinfo["results_dir"] = tmp_path
    return folderinfo


@pytest.fixture
def extract_cfg():
    cfg = {}
    cfg["do_permtest"] = False
    cfg["do_anova"] = True
    cfg["permutation_number"] = 100
    cfg["PCA_n_components"] = 3
    cfg["PCA_custom_scatter_PCs"] = ""
    cfg["PCA_save_3D_video"] = False
    cfg["PCA_bins"] = ""
    cfg["stats_threshold"] = 0.05
    cfg["plot_SE"] = False
    cfg["color_palette"] = "viridis"
    cfg["dont_show_plots"] = True
    cfg["legend_outside"] = True
    cfg["which_leg"] = "left"
    cfg["anova_design"] = "RM ANOVA"
    cfg["permutation_number"] = 100
    # NOTE - PCA & stats lists MUST be kept in this order
    # (otherwise PCA.Info & Stats.txt wont be equivalent to TRUE_DATA's)
    # => it's this order because it resulted from group_gui input (and thus corresponds to the checkbox-order of the features window)
    cfg["PCA_variables"] = [
        "Knee y",
        "Ankle y",
        "Hind paw tao y",
        "Ankle Angle",
        "Knee Angle",
        "Hip Angle",
    ]
    cfg["stats_variables"] = cfg["PCA_variables"]
    return cfg


# ..............................  RUN - ONE APPROVAL TEST  .............................


@pytest.mark.filterwarnings("ignore:Epsilon values")
@pytest.mark.slow
def test_group_approval(extract_true_dir, extract_folderinfo, extract_cfg):

    # ...........................  1) RUN GROUP GAITA  .................................
    group(extract_folderinfo, extract_cfg)

    # ......................  2) TEST EQUIVALENCE OF GROUP DFs  ........................
    # load true dfs from xlsx files
    true_av_df = pd.read_excel(
        os.path.join(extract_true_dir, "25 mm - Grand Average Group Stepcycles.xlsx")
    )
    true_std_df = pd.read_excel(
        os.path.join(
            extract_true_dir, "25 mm - Grand Standard Deviation Group Stepcycles.xlsx"
        )
    )
    test_av_df = pd.read_excel(
        os.path.join(
            extract_folderinfo["results_dir"],
            "25 mm - Grand Average Group Stepcycles.xlsx",
        )
    )
    test_std_df = pd.read_excel(
        os.path.join(
            extract_folderinfo["results_dir"],
            "25 mm - Grand Standard Deviation Group Stepcycles.xlsx",
        )
    )
    # finally assert equivalence of df-pairs
    pdt.assert_frame_equal(test_av_df, true_av_df)
    pdt.assert_frame_equal(test_std_df, true_std_df)

    # .......................  3) TEST EQUIVALENCE OF PCA DF  ..........................
    true_pca_df = pd.read_excel(os.path.join(extract_true_dir, "PCA Info.xlsx"))
    test_pca_df = pd.read_excel(
        os.path.join(extract_folderinfo["results_dir"], "PCA Info.xlsx")
    )
    pdt.assert_frame_equal(test_pca_df, true_pca_df)

    # ......................  4) TEST EQUIVALENCE OF STATS.TXT  ........................
    shallow = False  # if True compares only the metadata, not the contents!
    match, mismatch, errors = filecmp.cmpfiles(
        extract_true_dir,
        extract_folderinfo["results_dir"],
        [STATS_TXT_FILENAME],
        shallow,
    )
    assert match == [STATS_TXT_FILENAME]



from autogaita.universal3D.universal3D_datafile_preparation import (
    clean_a_file,
    rename_a_column,
)
from hypothesis import given, strategies as st
import pandas as pd


# %%..............................  preparation  .......................................


# property test for removing a string from column names via clean a file
column_names = st.lists(st.text(min_size=1), min_size=1, max_size=10, unique=True)
strings_to_remove = st.text(min_size=1)


@given(df_columns=column_names, string_to_remove=strings_to_remove)
def test_clean_a_file(df_columns, string_to_remove):
    df = pd.DataFrame(columns=df_columns)
    cleaned_df = clean_a_file(df, string_to_remove)
    for col in df_columns:
        if string_to_remove in col:
            assert col.replace(string_to_remove, "") in cleaned_df.columns
        else:
            assert col in cleaned_df.columns


# Property test for renaming a given column according to 3 possible cases
candidate_side_ids = st.sampled_from(
    ["l", "L", "left", "LEFT", "Left", "r", "R", "right", "RIGHT", "Right"]
)
candidate_coord_ids = st.sampled_from(["x", "y", "z", "X", "Y", "Z"])
candidate_joint_ids = st.text(min_size=2)
separators = st.sampled_from(["_", "-", ":", "."])


@given(
    candidate_side_id=candidate_side_ids,
    candidate_coord_id=candidate_coord_ids,
    candidate_joint_id=candidate_joint_ids,
    separator=separators,
)
def test_rename_a_column(
    candidate_side_id, candidate_coord_id, candidate_joint_id, separator
):
    # only run main test below if separator is not the joint id anywhere
    # => we require separators to be unique in column strings
    if separator in candidate_joint_id:
        return
    # preparation - what output we are expecting
    expected_side = "left" if candidate_side_id.lower() in ["l", "left"] else "right"
    expected_coord = candidate_coord_id.capitalize()
    # test 1 - column is side-specific
    col_string = separator.join(
        [candidate_side_id, candidate_coord_id, candidate_joint_id]
    )
    result = rename_a_column(col_string, separator)
    assert isinstance(result, str)
    assert (
        result == f"{candidate_joint_id}, {expected_side} {expected_coord.capitalize()}"
    )
    # test 2 - column is central
    col_string = separator.join([candidate_coord_id, candidate_joint_id])
    result = rename_a_column(col_string, separator)
    assert result == f"{candidate_joint_id} {expected_coord.capitalize()}"
    # test 3 - column is not a coordinate column
    col_string = separator.join([candidate_joint_id])
    result = rename_a_column(col_string, separator)
    assert result == col_string
